"""Classes to create binary data policies for positive and negative samples. Used by the :code:LocalClassifierPerNode."""
from abc import ABC

import networkx as nx
import numpy as np

from hiclass.data import NODE_TYPE


class Policy(ABC):
    """
    Abstract class used for all policies.

    Every policy should implement positive_samples and negative_samples methods.
    """

    def __init__(self, label_graph: nx.DiGraph, labels: np.ndarray):
        """
        Initialize a Policy with the required data.

        Parameters
        ----------
        label_graph : nx.DiGraph
            Graph which is used for inferring label relationships.
        labels : np.array
            Labels which will be assigned to the different samples.
            Has to be 1d array of integers.
        """
        self.graph = label_graph
        self.data = labels

    def positive_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all positive labels corresponding to the given label.

        Parameters
        ----------
        label : int or str
            Label for which the positive samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        raise NotImplementedError

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        raise NotImplementedError

    def _get_descendants(self, label: NODE_TYPE, inclusive: bool = True):
        """
        Gather all descendants for a given label.

        Parameters
        ----------
        label : int or str
            Label for which the descendants should be obtained.
        inclusive : bool, default=True
            True if the given label should be included in the list of descendants.

        Returns
        -------
        descendants : set
            Set of descendants for a given label.
        """
        descendants = set()
        if inclusive:
            descendants.add(label)
        for labels in nx.dfs_successors(self.graph, label).values():
            descendants.update(labels)
        return descendants

    def _get_siblings(self, label: NODE_TYPE):
        """
        Gather all siblings for a given label.

        Parameters
        ----------
        label : int or str
            Label for which the siblings should be obtained.

        Returns
        -------
        siblings : set
            Set of siblings for a given label.
        """
        parents = self.graph.predecessors(label)
        parents = list(parents)
        siblings = set()
        for parent in parents:
            siblings.update(self.graph.successors(parent))
        if label in siblings:
            siblings.remove(label)
        return siblings


class ExclusivePolicy(Policy):
    """Represent the exclusive policy of the referenced paper."""

    def positive_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all positive labels corresponding to the given label.

        This only includes the given label.

        Parameters
        ----------
        label : int or str
            Label for which the positive samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        return np.isin(self.data, label)

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        This includes all labels except the given label.

        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        negative_classes = np.logical_not(self.positive_samples(label))
        return negative_classes


class LessExclusivePolicy(ExclusivePolicy):
    """Represent the less exclusive policy of the referenced paper."""

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        This includes all labels except the given label and its children.

        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        negative_classes = set(self.graph.nodes)
        descendants = self._get_descendants(label)
        negative_classes.difference_update(descendants)
        return np.isin(self.data, list(negative_classes))


class ExclusiveSiblingsPolicy(ExclusivePolicy):
    """
    Represent the exclusive siblings policy of the referenced paper.

    Siblings are here defined as all classes that have the same distance to the root node (since DAGs are supported).
    """

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        This includes all labels that have the same distance from the root as the given label.
        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        siblings = self._get_siblings(label)
        return np.isin(self.data, list(siblings))


class InclusivePolicy(Policy):
    """Represent the inclusive policy of the referenced paper."""

    def positive_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all positive labels corresponding to the given label.

        This includes the label itself and all its descendants.

        Parameters
        ----------
        label : int or str
            Label for which the positive samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        descendants = self._get_descendants(label)
        return np.isin(self.data, list(descendants))

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        This includes all labels except the given label, its descendants and ancestors.

        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        negative_classes = set(self.graph.nodes)
        descendants = self._get_descendants(label)
        negative_classes.difference_update(descendants)
        ancestors = nx.ancestors(self.graph, label)
        negative_classes.difference_update(ancestors)
        return np.isin(self.data, list(negative_classes))


class SiblingsPolicy(InclusivePolicy):
    """
    Represent the siblings policy of the referenced paper.

    Siblings are defined here as all classes that have the same distance to the root node (since DAGs are supported).
    """

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        This includes all labels that have the same distance from the root as the given label, as well as all their
        children. In DAGs, this can include classes of the positive sample.

        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        siblings = self._get_siblings(label)
        negative_classes = set()
        for sibling in siblings:
            negative_classes.update(self._get_descendants(sibling))
        return np.isin(self.data, list(negative_classes))


class LessInclusivePolicy(InclusivePolicy):
    """Represent the less inclusive policy of the referenced paper."""

    def negative_samples(self, label: NODE_TYPE) -> np.ndarray:
        """
        Gather all negative labels corresponding to the given label.

        This includes all labels except the given label and its children.

        Parameters
        ----------
        label : int or str
            Label for which the negative samples should be searched.

        Returns
        -------
        positive_samples : np.array
            A mask for which labels are included (True) and which are not.
        """
        negative_classes = np.logical_not(self.positive_samples(label))
        return negative_classes


IMPLEMENTED_POLICIES = {
    "exclusive": ExclusivePolicy,
    "less_exclusive": LessExclusivePolicy,
    "exclusive_siblings": ExclusiveSiblingsPolicy,
    "inclusive": InclusivePolicy,
    "siblings": SiblingsPolicy,
    "less_inclusive": LessInclusivePolicy,
}



"""Helper functions for data manipulation and graph creation."""
from typing import Union, Tuple
import itertools

import networkx as nx
import numpy as np

PLACEHOLDER_LABEL_NUMERIC = -1
PLACEHOLDER_LABEL_CAT = ""
NODE_TYPE = Union[int, str, np.str_]
CATEGORICAL_ROOT_NODE = "HICLASS::__ROOT__"
SEPARATOR = "HICLAS::__SEPARATOR__"

# Boolean, unsigned integer, signed integer, float, complex.
_NUMERIC_KINDS = set("buif")


def graph_from_edge_pairs(
    file: str, delimiter: str = ",", skip_header: int = 1
) -> nx.DiGraph:
    """
    Create a DAG from a file containing (parent, child) pairs.

    Parameters
    ----------
    file : str
        File containing the edge pairs.
    delimiter : str, default=','
        The delimiter of the file.
    skip_header : int, default=1
        The number of rows header rows that should be skipped.

    Returns
    -------
    graph : nx.DiGraph
        The graph with all corresponding edges and their nodes.
    """
    edges = np.genfromtxt(file, delimiter=delimiter, skip_header=skip_header)
    return nx.from_edgelist(edges, create_using=nx.DiGraph)


def graph_from_hierarchical_labels(
    data: np.ndarray, placeholder: Union[str, int, None] = None
) -> nx.DiGraph:
    """
    Construct a DAG from hierarchical labels.

    In the case that multiple root nodes are found, a new root node is inserted and all previous root nodes are
    connected to it.

    Parameters
    ----------
    data : np.array
        Hierarchical labels, formatted to be (row, col). The columns should be ordered
        from least specific to most specific class. If some columns are invalid (i.e. there are columns with a
        number of labels lower than the number of columns), then they should be marked by a placeholder.
    placeholder : str or int, default=None
        Value for non-existent nodes in the data. Has to match data type of :code:`data`

    Returns
    -------
    graph : DiGraph
        The graph with all corresponding edges and their nodes.
    """
    g = nx.DiGraph()
    nodes = np.unique(data)
    if placeholder in nodes:
        nodes = list(nodes)
        nodes.remove(placeholder)
    g.add_nodes_from(nodes)

    if data.ndim > 1:
        roots = np.unique(data[:, 0])
    else:
        roots = np.unique(data)
    if len(roots) > 1:
        new_root = max(nodes)
        new_root = new_root + 1 if is_numeric_label(new_root) else CATEGORICAL_ROOT_NODE

        g.add_edges_from(zip(itertools.repeat(new_root), roots))

    col = 0
    if data.ndim > 1:
        while col + 2 <= data.shape[1]:
            valid_edges = np.unique(data[:, col : col + 2], axis=0)
            placeholders = np.any(valid_edges == placeholder, axis=1)
            valid_edges = valid_edges[np.invert(placeholders)]
            g.add_edges_from(valid_edges)
            col += 1
    return g


def convert_categorical_labels_to_numeric(
    data: np.ndarray, placeholder_label: str = PLACEHOLDER_LABEL_CAT
) -> Tuple[np.ndarray, dict]:
    """
    Take a string array and convert the values to enumerated integers.

    Parameters
    ----------
    data : np.array
        Data to be processed.
    placeholder_label : str
        label representing no present class.

    Returns
    -------
    conversion : np.array
        The converted integer array.
    mapping : dict
         Mapping from the original value to the new enumeration.
    """
    categories, indices, mapping = np.unique(
        data, return_index=True, return_inverse=True
    )
    label_to_int = dict(zip(categories, mapping[indices]))
    # map all undefined_labels to -1, implying an undefined output and remove it from the mapping
    if placeholder_label in categories:
        mapping[mapping == label_to_int[placeholder_label]] = PLACEHOLDER_LABEL_NUMERIC
        label_to_int[placeholder_label] = PLACEHOLDER_LABEL_NUMERIC
    return mapping.reshape(data.shape), label_to_int


def find_root(graph: nx.DiGraph) -> NODE_TYPE:
    """Take a graph and return one of its root nodes.

    Parameters
    ----------
    graph : nx.DiGraph
        Graph for which the root should be found.

    Returns
    -------
    root : int or str
        The label of the root node.
    """
    current_node = list(graph.nodes)[0]
    successors = list(graph.predecessors(current_node))
    while len(successors) != 0:
        current_node = successors[0]
        successors = list(graph.predecessors(current_node))
    return current_node


def find_max_depth(graph: nx.DiGraph, root: NODE_TYPE) -> int:
    """Find the maximum depth of a DAG given a node as root.

    Parameters
    ----------
    graph : nx.DiGraph
        Graph for which the root should be found.
    root : int or str
        Node from which the distance should be measured.

    Returns
    -------
    max_depth : int
        The number of nodes in the furthest path starting from the root node. A graph that only has the root would
        return 1.
    """
    # Queue root node and initialize max_depth
    queue = [root]
    max_depth = 0

    while True:

        # node_count indicates the number of nodes at current level
        node_count = len(queue)
        if node_count == 0:
            return max_depth

        max_depth += 1

        # dequeue all nodes from current level
        # and queue nodes from next level
        while node_count > 0:
            node = queue.pop(0)
            for successor in list(graph.successors(node)):
                queue.append(successor)
            node_count -= 1


def minimal_graph_depth(graph: nx.DiGraph) -> int:
    """
    Calculate the minimal depth in which all nodes can be hit.

    Parameters
    ----------
    graph : nx.DiGraph
        Graph to be analyzed.

    Returns
    -------
    depth : int
        The minimal depth.
    """
    max_level = 1
    for level in minimal_per_node_depth(graph).values():
        max_level = max(level, max_level)
    return max_level


def minimal_per_node_depth(graph: nx.DiGraph) -> dict:
    """
    Calculate the minimal depth which is needed to hit a node, for all nodes.

    Parameters
    ----------
    graph : nx.DiGraph
        Graph to be analyzed.

    Returns
    -------
    node_depth : dict
        A mapping for node : depth
    """
    root = find_root(graph)
    node_depth = {root: 1}
    for parent, successors in nx.bfs_successors(graph, root):
        parent_level = node_depth[parent]
        for node in successors:
            if node not in node_depth.keys():
                node_depth[node] = parent_level + 1
    return node_depth


def is_numeric_label(array: np.ndarray) -> bool:
    """Determine whether an array has a numerical label format.

    Supported formats are booleans, unsigned integers, signed integers and floats.

    Parameters
    ----------
    array : np.array
        The array to check.

    Returns
    -------
    result : bool
        True if the array is has a supported format, False otherwise
    """
    return np.asarray(array).dtype.kind in _NUMERIC_KINDS


def flatten_labels(labels: np.ndarray, placeholder_label: int = -1) -> np.ndarray:
    """Flattens hierarchical labels to only the most specific label per entry.

    Expects the most specific label to be on the rightmost side of the label array for every entry.

    Parameters
    ----------
    labels : np.array
        2d label matrix, formatted to be row, column.
    placeholder_label : int or str
        value describing an undefined label in order to support uneven hierarchies and labels.

    Returns
    -------
    new_labels: np.array
        1d array of the most specific label for each row.
    """
    if len(labels.shape) < 2:
        return labels
    num_labels = len(labels)
    new_labels = np.zeros(num_labels, dtype=labels.dtype)
    for i, hierarchical_label in enumerate(labels):
        for label in hierarchical_label[::-1]:
            if label != placeholder_label:
                new_labels[i] = label
                break
    return new_labels



"""Shared code for all classifiers."""
import abc
import logging
import multiprocessing
import sys
from itertools import starmap
from typing import Any, Callable, Iterator, List, Optional, Tuple, Type, Union

import networkx as nx
import numpy as np
from sklearn.base import BaseEstimator, clone, is_classifier

import hiclass.data as data_utils
from hiclass import Policies
from hiclass.data import (
    CATEGORICAL_ROOT_NODE,
    NODE_TYPE,
    SEPARATOR,
    minimal_per_node_depth,
)


class Classifier(abc.ABC):
    """Abstract class for the local hierarchical classifiers.

    Offers mostly utility methods and common data initialization. This includes getting and setting classifiers on nodes
    as well as calculating and accepting the class hierarchy.
    """

    def __init__(
        self,
        n_jobs: int = 1,
        verbose: int = 0,
        local_classifier: Any = BaseEstimator(),
        hierarchy: Optional[nx.DiGraph] = None,
        unique_taxonomy: bool = True,
    ):
        """
        Initialize a classifier.

        Parameters
        ----------
        n_jobs : int, default=1
            The number of jobs to run in parallel. Only :code:`fit` is parallelized.
        verbose : int, default=0
            Controls the verbosity when fitting and predicting.
            See https://verboselogs.readthedocs.io/en/latest/readme.html#overview-of-logging-levels
            for more information.
        local_classifier : BaseEstimator instance
            The local_classifier used to create the collection of local classifiers. Needs to have fit, predict and
            clone methods.
        hierarchy : nx.DiGraph, default=None
            Label hierarchy used in prediction and fitting. If None, it will be inferred during training.
        unique_taxonomy : bool, default=True
            True if the elements in the hierarchy have unique names, otherwise it can have unexpected behaviour.
            For example, a->b->c and d->b->e could have different meanings for b, so in that case unique_taxonomy
            should be set to false.
        """
        self.n_jobs = n_jobs
        self.log = logging.getLogger(__name__)
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(verbose)
        sh.setFormatter(
            logging.Formatter("[%(levelname)s]\t [%(name)s]\t : %(message)s")
        )
        self.log.addHandler(sh)
        self.local_classifier = local_classifier
        self.hierarchy = hierarchy
        self.unique_taxonomy = unique_taxonomy
        self.placeholder_label = None
        self.max_depth = None
        if self.hierarchy is not None:
            self._initialize_hierarchy_information()
        super().__init__()

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        placeholder_label: Optional[NODE_TYPE] = None,
        replace_classifiers: bool = True,
    ):
        """
        Fits all classifiers.

        Needs to be subclassed by other classifiers as it only offers hierarchy methods.

        Parameters
        ----------
        X : np.array of shape (n_samples, n_features)
            The training input samples.
        Y : np.array of shape (n_samples, n_levels)
            The hierarchical labels.
        placeholder_label : int or str, default=None
            Label that corresponds to "no label available for this data point". Defaults will be used if not passed.
        replace_classifiers : bool, default=True
            Turns on (True) the replacement of a local classifier with a constant classifier when trained on only
            a single unique class.
        """
        if placeholder_label is not None:
            self.placeholder_label = placeholder_label
        if not self.unique_taxonomy:
            Y = self._make_unique(Y)
        if self.hierarchy is None:
            self.hierarchy = data_utils.graph_from_hierarchical_labels(
                Y, self.placeholder_label
            )
            self._initialize_hierarchy_information()

    def predict(self, X: np.ndarray, threshold: float = 0):
        """
        Predict classes for the given data.

        Hierarchical labels are returned. If threshold is specified, prediction can end early.

        Parameters
        ----------
        X : np.array of shape (n_samples, n_features)
            The input samples.
        threshold : float, default=None
            Minimum confidence score to continue prediction for children nodes.
        Returns
        -------
        y : np.array of shape (n_samples, n_levels)
            The predicted classes.
        """
        final_prediction = self._init_prediction(X)
        self._specialize_prediction(
            self.root,
            depth=0,
            data=X,
            threshold=threshold,
            final_prediction=final_prediction,
            selected_rows=np.ones(np.shape(X)[0], dtype=np.bool),
        )
        final_prediction = self._delete(CATEGORICAL_ROOT_NODE, final_prediction)
        if not self.unique_taxonomy:
            final_prediction = self._remove_separator(final_prediction)
        return final_prediction

    def _remove_separator(self, prediction):
        for i in range(np.shape(prediction)[0]):
            for j in range(1, np.shape(prediction)[1]):
                prediction[i, j] = prediction[i, j].split(SEPARATOR)[-1]
        return prediction

    def _make_unique(self, Y):
        new_Y = []
        for i in range(np.shape(Y)[0]):
            row = [Y[i, 0]]
            for j in range(1, np.shape(Y)[1]):
                row.append(row[-1] + SEPARATOR + Y[i, j])
            new_Y.append(np.asarray(row, dtype=np.str_))
        return np.array(new_Y)

    def _delete(self, character, prediction):
        final_prediction = []
        for i in range(np.shape(prediction)[0]):
            row = prediction[i]
            row = np.array(row[row != character]).flatten()
            final_prediction.append(row)
        final_prediction = np.asarray(final_prediction, dtype=object)
        return final_prediction

    def _init_prediction(self, X: np.ndarray):
        return np.full(
            (np.shape(X)[0], self.max_depth),
            self.placeholder_label,
            dtype=object if self.categorical_hierarchy else np.int32,
        )

    def _fit_classifier(self, fit_fct: Callable, fit_args: Iterator[Tuple]) -> None:
        if self.n_jobs > 1:
            with multiprocessing.Pool(self.n_jobs) as pool:
                results = pool.starmap(fit_fct, fit_args)
        else:
            results = list(starmap(fit_fct, fit_args))

        for descriptor, classifier, warning in results:
            self.set_classifier(descriptor, classifier)
            if warning:
                self.log.warning(warning)

    def _specialize_prediction(
        self,
        local_identifier: NODE_TYPE,
        depth: int,
        data: np.ndarray,
        threshold: float,
        final_prediction: np.ndarray,
        selected_rows: np.ndarray,
    ) -> None:
        nodes_to_predict = self._get_nodes_to_predict(local_identifier)
        additional_prediction_args = self._choose_required_prediction_args(
            local_identifier, nodes_to_predict, selected_rows
        )

        if self._stopping_criterion(nodes_to_predict):
            return

        prediction = self._create_prediction(data, *additional_prediction_args)
        most_likely_labels, rows_under_threshold = self._analyze_prediction(
            prediction, threshold
        )

        self._continue_prediction(
            nodes_to_predict,
            selected_rows,
            most_likely_labels,
            rows_under_threshold,
            depth + 1,
            data,
            threshold,
            final_prediction,
        )

    def _get_nodes_to_predict(self, local_identifier: NODE_TYPE) -> List[NODE_TYPE]:
        raise NotImplementedError

    def _choose_required_prediction_args(
        self,
        local_identifier: NODE_TYPE,
        nodes_to_predict: List[NODE_TYPE],
        selected_rows: np.ndarray,
    ) -> Tuple:
        raise NotImplementedError

    def _stopping_criterion(self, value: Any) -> bool:
        raise NotImplementedError

    def _create_prediction(self, data: np.ndarray, *args) -> np.ndarray:
        raise NotImplementedError

    def _predict_with_classifier(self, classifier: Any, data: np.ndarray):
        if hasattr(classifier, "predict_proba"):
            return classifier.predict_proba(data)
        else:
            self.log.warning(
                "Local classifier does not support .predict_proba(). Trying with .predict()"
            )
            return classifier.predict(data)

    @staticmethod
    def _analyze_prediction(
        predictions: np.ndarray, threshold: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        max_predictions = np.max(predictions, axis=0)
        most_likely_labels = np.argmax(predictions, axis=0)

        under_threshold = max_predictions < threshold
        return most_likely_labels, under_threshold

    def _continue_prediction(
        self,
        nodes_to_predict: List[NODE_TYPE],
        selected_rows: np.ndarray,
        most_likely_labels: np.ndarray,
        rows_under_threshold: np.ndarray,
        depth: int,
        data: np.ndarray,
        threshold: float,
        final_prediction: np.ndarray,
    ) -> None:
        raise NotImplementedError

    def _initialize_hierarchy_information(self) -> None:
        self.root = data_utils.find_root(self.hierarchy)
        if self.max_depth is None:
            self.max_depth = data_utils.find_max_depth(self.hierarchy, self.root)
        self.categorical_hierarchy = isinstance(self.root, str)
        if self.placeholder_label is None:
            if self.categorical_hierarchy:
                self.placeholder_label = data_utils.PLACEHOLDER_LABEL_CAT
            else:
                self.placeholder_label = data_utils.PLACEHOLDER_LABEL_NUMERIC
        self._initialize_classifiers()

    def _initialize_classifiers(self) -> None:
        depth_per_node = minimal_per_node_depth(self.hierarchy)
        max_considered_depth = max(depth_per_node.values())
        self.classes_ = [[] for _ in range(max_considered_depth)]
        for node, level in depth_per_node.items():
            self.classes_[level - 1].append(node)
        for level in range(len(self.classes_)):
            self.classes_[level] = np.array(self.classes_[level])

    def _copy_classifier(self):
        if is_classifier(self.local_classifier):
            return clone(self.local_classifier)
        else:
            try:
                return self.local_classifier.clone()
            except AttributeError:
                error_msg = (
                    "Assigned local classifier is not made by sklearn. Trying to copy by calling .clone()"
                    " failed."
                )
                self.log.error(error_msg)
                raise AttributeError(error_msg)

    def _copy_and_assign_classifier(self, node: NODE_TYPE) -> None:
        classifier = self._copy_classifier()
        self.hierarchy.nodes[node]["classifier"] = classifier

    def get_classifier(self, descriptor: Any) -> Any:
        """
        Return the local classifier associated to the given descriptor.

        Parameters
        ----------
        descriptor : Any
            The descriptor for which the local classifier should be returned.

        Returns
        -------
        classifier : Any
            The local classifier that is assigned to the given descriptor.
        """
        raise NotImplementedError

    def set_classifier(self, descriptor: Any, classifier: Any) -> None:
        """
        Set the local classifier for a given descriptor.

        Parameters
        ----------
        descriptor : Any
            The descriptor for which the local classifier should be set.
        classifier : Any
            The local classifier that will be assigned to the given descriptor.
        """
        raise NotImplementedError


class NodeClassifier(Classifier, abc.ABC):
    """Abstract class for classifiers that have their local classifiers linked to nodes."""

    def __init__(
        self,
        n_jobs: int = 1,
        verbose: int = 0,
        local_classifier: Any = BaseEstimator(),
        hierarchy: Optional[nx.DiGraph] = None,
        unique_taxonomy: bool = True,
        policy: Union[str, Type[Policies.Policy]] = "siblings",
    ):
        """
        Initialize a classifier.

        Extends the superclass by adding a data policy used in fitting.

        Parameters
        ----------
        n_jobs : int, default=1
            The number of jobs to run in parallel. Only :code:`fit` is parallelized.
        verbose : int, default=0
            Controls the verbosity when fitting and predicting.
            See https://verboselogs.readthedocs.io/en/latest/readme.html#overview-of-logging-levels
            for more information.
        local_classifier : BaseEstimator instance
            The local classifier used to create the collection of local classifiers. Needs to have fit, predict and
            clone methods.
        hierarchy : nx.DiGraph, default=None
            Label hierarchy used in prediction and fitting. If None, it will be inferred during training.
        unique_taxonomy : bool, default=True
            True if the elements in the hierarchy have unique names, otherwise it can have unexpected behaviour.
            For example, a->b->c and d->b->e could have different meanings for b, so in that case unique_taxonomy
            should be set to false.
        policy : Policy, default="siblings"
            Rules for defining positive and negative training samples.
        """
        self.policy = policy
        super().__init__(
            n_jobs=n_jobs,
            verbose=verbose,
            local_classifier=local_classifier,
            hierarchy=hierarchy,
            unique_taxonomy=unique_taxonomy,
        )

    def _specialize_prediction(
        self,
        local_identifier: NODE_TYPE,
        depth: int,
        data: np.ndarray,
        threshold: float,
        final_prediction: np.ndarray,
        selected_rows: np.ndarray,
    ) -> None:
        final_prediction[selected_rows, depth] = local_identifier
        super(NodeClassifier, self)._specialize_prediction(
            local_identifier, depth, data, threshold, final_prediction, selected_rows
        )

    def _get_nodes_to_predict(self, local_identifier: NODE_TYPE) -> List[NODE_TYPE]:
        return list(self.hierarchy.successors(local_identifier))

    def _choose_required_prediction_args(
        self,
        local_identifier: NODE_TYPE,
        nodes_to_predict: List[NODE_TYPE],
        selected_rows: np.ndarray,
    ) -> Tuple[np.ndarray, NODE_TYPE]:
        return selected_rows, local_identifier

    def _stopping_criterion(self, nodes_to_predict: List[NODE_TYPE]) -> bool:
        return len(nodes_to_predict) == 0

    def _continue_prediction(
        self,
        nodes_to_predict: List[NODE_TYPE],
        selected_rows: np.ndarray,
        most_likely_labels: np.ndarray,
        rows_under_threshold: np.ndarray,
        depth: int,
        data: np.ndarray,
        threshold: float,
        final_prediction: np.ndarray,
    ) -> None:
        if np.all(rows_under_threshold):
            return

        for child in nodes_to_predict:
            corresponding_data = np.copy(selected_rows)
            corresponding_data[selected_rows] = np.logical_and(
                most_likely_labels == nodes_to_predict.index(child),
                np.logical_not(rows_under_threshold),
            )
            if any(corresponding_data):
                self._specialize_prediction(
                    child,
                    depth,
                    data,
                    threshold,
                    final_prediction,
                    corresponding_data,
                )

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        placeholder_label: Optional[NODE_TYPE] = None,
        replace_classifiers: bool = True,
    ) -> None:
        """
        Fit all classifiers.

        Extends superclass method and needs to be subclassed by other classifiers. Adds label flattening and policy
        initialization.

        Parameters
        ----------
        X : np.array of shape (n_samples, n_features)
            The training input samples.
        Y : np.array of shape (n_samples, n_levels)
            The hierarchical labels.
        placeholder_label : int or str, default=None
            Label that corresponds to "no label available for this data point". Defaults will be used if not passed.
        replace_classifiers : bool, default=True
            Turns on (True) the replacement of a local classifier with a constant classifier when trained on only
            a single unique class.
        """
        super().fit(X, Y, placeholder_label, replace_classifiers)

        if not self.unique_taxonomy:
            Y = self._make_unique(Y)

        labels = data_utils.flatten_labels(Y, self.placeholder_label)
        if type(self.policy) == str:
            try:
                self.policy = Policies.IMPLEMENTED_POLICIES[self.policy.lower()](
                    self.hierarchy, labels
                )
            except KeyError:
                raise_error_and_log(
                    KeyError,
                    self.log,
                    f"Policy {self.policy} not implemented. Available policies are:\n"
                    f"{list(Policies.IMPLEMENTED_POLICIES.keys())}",
                )
        else:
            self.policy = self.policy(self.hierarchy, labels)

    def _initialize_classifiers(self) -> None:
        super()._initialize_classifiers()
        for node in self._get_nodes_with_classifier():
            self._copy_and_assign_classifier(node)

    def _get_nodes_with_classifier(self) -> List[NODE_TYPE]:
        raise NotImplementedError

    def _get_node(self, label: NODE_TYPE):
        if type(label) == str or (
            hasattr(label, "dtype") and label.dtype.type is np.str_
        ):
            if self.categorical_hierarchy:
                return self.hierarchy.nodes[label]
            raise_error_and_log(
                ValueError,
                self.log,
                "String label was given, but label hierarchy is numeric.",
            )
        else:
            if self.categorical_hierarchy:
                raise_error_and_log(
                    ValueError,
                    self.log,
                    "Numeric label was given, but label hierarchy is categorical.",
                )
            else:
                return self.hierarchy.nodes[label]

    def get_classifier(self, descriptor: NODE_TYPE) -> Any:
        """
        Return the local classifier associated to the given label.

        Parameters
        ----------
        descriptor : int or str
            The label for which the local classifier should be returned.

        Returns
        -------
        classifier : Any
            The local classifier that is assigned to the given label.
        """
        return self._get_node(descriptor)["classifier"]

    def set_classifier(self, descriptor: NODE_TYPE, classifier: Any) -> None:
        """
        Set the local classifier for a given label.

        Parameters
        ----------
        descriptor : int or str
            The label for which the local classifier should be set.
        classifier : Any
            The local classifier that will be assigned to the given label.
        """
        self._get_node(descriptor)["classifier"] = classifier


class ConstantClassifier:
    """A classifier that returns 1 for a specified label for all samples during prediction."""

    def __init__(self, class_to_predict: int, num_classes: int) -> None:
        """
        Initialize the classifier.

        Parameters
        ----------
        class_to_predict : int
            The index of the label that should be predicted with a probability of one. Needs to be within `num_classes`
        num_classes : int
            The amount of labels that should be predicted by the classifier.
        """
        self.value = np.zeros(num_classes, dtype=int)
        self.value[class_to_predict] = 1

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        """
        Predict X with previously set parameters.

        Parameters
        ----------
        X : np.ndarray of shape(n_samples, ...)
            Data that should be predicted. Only the number of samples matters.

        Returns
        -------
        output : np.ndarray
            1 for the previously set label and 0 for all others for all samples in X.
        """
        return np.vstack([self.value] * np.shape(X)[0])


def fit_or_replace(
    replace_classifier: bool,
    classifier: Any,
    data: np.ndarray,
    labels: np.ndarray,
    out_dim: int,
):
    """
    Fit a given local classifier on data and labels and replace it if there is only a single label to predict.

    Parameters
    ----------
    replace_classifier : bool
        Only possibly replace the classifier if this is True. Otherwise always try to fit the classifier.
    classifier : Any
        Local classifier to be fitted.
    data : np.ndarray
        Data used for fitting.
    labels : np.ndarray
        Classes used for fitting.
    out_dim : int
        Number of columns that the classifier should predict.

    Returns
    -------
    classifier : Any,
        Either the fitted classifier or a ConstantClassifier.
    warning_msg : str or None
        None if a fitted classifier is returned, a warning message otherwise.
    """
    unique_labels = np.unique(labels)
    warning = None
    if replace_classifier and len(unique_labels) == 1:
        warning = "One or more classifiers could not be fitted since all data points have the same label."
        warning = warning + ' These were replaced with a "constant classifier".'
        classifier = ConstantClassifier(unique_labels[0], out_dim)
    else:
        classifier.fit(data, labels)
    return classifier, warning


class DuplicateFilter:
    """A filter that removes duplicate messages from logging."""

    def __init__(self):
        """Initialize filter."""
        self.msgs = set()

    def filter(self, record: Any) -> bool:
        """
        Filter messages.

        Parameters
        ----------
        record : logged message
            Text to be filtered.

        Returns
        -------
        rv : bool
            True if message was not found in the filter, false otherwise.
        """
        rv = record.msg not in self.msgs
        self.msgs.add(record.msg)
        return rv


def raise_error_and_log(
    error: Type[Exception], logger: logging.Logger, msg: str
) -> None:
    """
    Log a message and throw it as an error.

    Parameters
    ----------
    error : Exception
        The type of error to throw.
    logger : logging.Logger
        The logger to use for logging.
    msg : str
        Message to throw and to log.
    """
    logger.error(msg)
    raise error(msg)



"""
Local classifier per level approach.

Numeric and string output labels are both handled.
"""
from itertools import repeat
from typing import Any, Tuple, Type, Union, List, Optional

import numpy as np

from hiclass.data import NODE_TYPE
from hiclass.Classifier import (
    Classifier,
    DuplicateFilter,
    fit_or_replace,
    ConstantClassifier,
    raise_error_and_log,
)


class DefaultProbability:
    """Defines probability calculations used to return probabilities for all local classifiers."""

    def predict_proba(self, classifiers: List[Any], X: np.ndarray) -> List:
        """
        Compute prediction probabilities.

        Parameters
        ----------
        classifiers : List of classifiers
            The local classifiers that should be used for the prediction
        X : np.array of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        probabilities : np.ndarray of shape (n_levels, n_samples)
            Prediction probabilities for all classes in each hierarchical level.
        """
        probabilities = []
        for classifier in classifiers:
            probabilities.append(classifier.predict_proba(X))
        return probabilities


IMPLEMENTED_PROBABILITIES = {"default": DefaultProbability}


def _fit_classifier(
    level: int,
    classifier: Any,
    level_nodes: np.array,
    data: np.ndarray,
    labels: np.ndarray,
    placeholder: NODE_TYPE,
    replace_classifier: bool,
) -> Tuple[int, Any, str]:
    """Fits a local classifier for a given level and data and metadata.

    Parameters
    ----------
    level : int
        The level that the classifier should be assigned to.
    classifier : Any
        The classifier that is fitted.
    level_nodes : np.array
        A list of nodes from the hierarchy that belong to the given `level`.
    data : np.array
        The data that is being used for training the classifier.
    labels : np.array
        The labels corresponding to the given data. Must have the same number of rows as `data`.
    placeholder : int or str
        A placeholder descriptor for non-existing labels in the `labels` argument.
    replace_classifier : bool
        Turns on (True) the replacement of classifiers with a constant classifier when trained on only
        a single unique label.

    Returns
    -------
    level : int
        Level for which the classifier was fitted.
    classifier : Any
        The classifier that was fitted for the given node.
    warning : str
        The warnings raised.
    """
    shifted_level = level - 1
    relevant_data = labels[:, shifted_level] != placeholder
    classes = np.zeros_like(relevant_data, dtype=int)

    mapping = dict(zip(level_nodes, range(0, len(level_nodes))))
    for node in level_nodes:
        positive_samples = labels[:, shifted_level] == node
        classes[positive_samples] = mapping[node]

    classifier, warning = fit_or_replace(
        replace_classifier,
        classifier,
        data[relevant_data],
        classes[relevant_data],
        len(mapping),
    )
    return level, classifier, warning


class LocalClassifierPerLevel(Classifier):
    """
    Assign local classifiers for each class hierarchy level.

    A local classifier per level is a local hierarchical classifier that fits one local multi-class classifier
    for each level of the hierarchy. In case of a DAG, nodes are assigned their highest possible level, with the root
    being the highest level.
    """

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        placeholder_label: Optional[NODE_TYPE] = None,
        replace_classifiers: bool = True,
    ) -> None:
        """
        Fit the local classifiers.

        Parameters
        ----------
        X : np.array of shape(n_samples, n_features)
            The training input samples.
        Y : np.array of shape (n_samples, n_levels)
            The hierarchical labels.
        placeholder_label : int or str, default=None
            Label that corresponds to "no label available for this data point". Defaults will be used if not passed.
        replace_classifiers : bool, default=True
            Turns on (True) the replacement of a local classifier with a constant classifier when trained on only
            a single unique class.
        """
        super().fit(X, Y, placeholder_label, replace_classifiers)

        if not self.unique_taxonomy:
            Y = self._make_unique(Y)

        duplicate_filter = DuplicateFilter()
        self.log.addFilter(duplicate_filter)  # removes possible duplicate warnings

        if all(Y[:, 0] == self.root):
            Y = Y[:, 1:]
        if any(Y[:, 0] == self.root):
            error_msg = (
                "Some labels contain the root node and some do not. Classifier can only be fit when "
                "labels are consistent."
            )
            raise_error_and_log(ValueError, self.log, error_msg)

        fit_args = zip(
            range(0, len(self.classes_)),
            self.classifiers,
            self.classes_,
            repeat(X),
            repeat(Y),
            repeat(placeholder_label),
            repeat(replace_classifiers),
        )

        next(fit_args)  # skip the root node
        self._fit_classifier(_fit_classifier, fit_args)
        self.classifiers[0] = ConstantClassifier(0, 1)

        self.log.removeFilter(duplicate_filter)  # delete duplicate filter

    def predict_proba(
        self, X: np.ndarray, algorithm: Union[str, Type[DefaultProbability]] = "default"
    ) -> List:
        """
        Compute prediction probabilities.

        Parameters
        ----------
        X : np.array of shape(n_samples, n_features)
            The input samples.
        algorithm : str or Probability
            The algorithm to use for calculating probabilities.

        Returns
        -------
        probabilities : np.ndarray of shape (n_levels, n_samples)
            Prediction probabilities for all classes in each hierarchical level.
        """
        if type(algorithm) == str:
            try:
                probability_algorithm = IMPLEMENTED_PROBABILITIES[algorithm.lower()]()
            except KeyError:
                error_msg = (
                    f"Probability algorithm {algorithm} not implemented. Available algorithms are:\n"
                    f"{list(IMPLEMENTED_PROBABILITIES.keys())}"
                )
                raise_error_and_log(KeyError, self.log, error_msg)
        else:
            probability_algorithm = algorithm
        return probability_algorithm.predict_proba(self.classifiers, X)

    def _ensure_possible_predictions(
        self, final_prediction: np.ndarray, predictions: np.ndarray, depth: int
    ) -> np.ndarray:
        """Zero prediction results that are not possible."""
        if depth == 0:
            return predictions
        previous_predictions = np.unique(final_prediction[:, depth - 1])
        corrected_predictions = np.zeros_like(predictions)
        possible_children = [
            list(self.hierarchy.successors(node)) for node in previous_predictions
        ]
        for previous_prediction, possible_predictions in zip(
            previous_predictions, possible_children
        ):
            relevant_rows = np.where(
                final_prediction[:, depth - 1] == previous_prediction
            )[0][:, None]
            possible_predictions = np.array(
                [
                    np.where(self.classes_[depth] == node)[0][0]
                    for node in possible_predictions
                ]
            )
            corrected_predictions[relevant_rows, possible_predictions] = predictions[
                relevant_rows, possible_predictions
            ]
        return corrected_predictions

    def _specialize_prediction(
        self,
        local_identifier: NODE_TYPE,
        depth: int,
        data: np.ndarray,
        threshold: float,
        final_prediction: np.ndarray,
        selected_rows: np.ndarray,
    ) -> None:
        local_identifier = depth
        if self._stopping_criterion(depth):
            return

        nodes_to_predict = self._get_nodes_to_predict(local_identifier)
        additional_prediction_args = self._choose_required_prediction_args(
            local_identifier, nodes_to_predict, selected_rows
        )

        predictions = self._create_prediction(data, *additional_prediction_args)
        predictions = self._ensure_possible_predictions(
            final_prediction, predictions, depth
        ).T

        most_likely_labels, rows_under_threshold = self._analyze_prediction(
            predictions, threshold
        )
        selected_rows = np.logical_and(
            selected_rows, np.logical_not(rows_under_threshold)
        )

        self._continue_prediction(
            nodes_to_predict,
            selected_rows,
            most_likely_labels,
            rows_under_threshold,
            depth,
            data,
            threshold,
            final_prediction,
        )

    def _get_nodes_to_predict(self, local_identifier: int) -> np.ndarray:
        return self.classes_[local_identifier]

    def _choose_required_prediction_args(
        self,
        local_identifier: NODE_TYPE,
        nodes_to_predict: np.ndarray,
        selected_rows: np.ndarray,
    ) -> Any:
        return [local_identifier]

    def _stopping_criterion(self, value: int) -> bool:
        return value >= len(self.classes_)

    def _create_prediction(self, data: np.ndarray, depth: int) -> np.ndarray:
        classifier = self.get_classifier(depth)
        return self._predict_with_classifier(classifier, data)

    def _continue_prediction(
        self,
        nodes_to_predict: np.ndarray,
        selected_rows: np.ndarray,
        most_likely_labels: np.ndarray,
        rows_under_threshold: np.ndarray,
        depth: int,
        data: np.ndarray,
        threshold: float,
        final_prediction: np.ndarray,
    ):
        if np.all(rows_under_threshold):
            return

        for label in nodes_to_predict:
            corresponding_data = np.copy(selected_rows)
            label_idx = np.where(nodes_to_predict == label)[0][0]
            corresponding_data[selected_rows] = (most_likely_labels == label_idx)[
                selected_rows
            ]
            final_prediction[corresponding_data, depth] = label

        self._specialize_prediction(
            depth + 1, depth + 1, data, threshold, final_prediction, selected_rows
        )

    def _initialize_classifiers(self) -> None:
        super()._initialize_classifiers()
        self.classifiers = [self._copy_classifier() for _ in self.classes_]

    def get_classifier(self, descriptor: int) -> Any:
        """
        Return the local classifier associated to the given hierarchy level.

        Raise IndexError if the level is invalid.

        Parameters
        ----------
        descriptor : int or str
            the descriptor for which the local classifier should be returned.

        Returns
        -------
        classifier : Any
            The local classifier that is assigned to the given descriptor.
        """
        if descriptor < 0 or descriptor >= len(self.classes_):
            raise IndexError("Hierarchy level does not exist.")
        return self.classifiers[descriptor]

    def set_classifier(self, descriptor: int, classifier: Any) -> None:
        """
        Set the local classifier for a given hierarchy level.

        Parameters
        ----------
        descriptor : int or str
            the descriptor for which the local classifier should be set.
        classifier : Any
            The local classifier that will be assigned to the given label.
        """
        self.classifiers[descriptor] = classifier



"""
Local classifier per node approach.

Numeric and string output labels are both handled.
"""
import logging
from copy import deepcopy

import networkx as nx
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted

from hiclass.ConstantClassifier import ConstantClassifier
from hiclass import BinaryPolicy


class LocalClassifierPerNode(BaseEstimator):
    """
    Assign local classifiers to each node of the graph, except the root node.

    A local classifier per node is a local hierarchical classifier that fits one local binary classifier
    for each node of the class hierarchy, except for the root node.
    """

    def __init__(
        self,
        local_classifier: BaseEstimator = None,
        binary_policy: str = "siblings",
        verbose: int = 0,
        edge_list: str = None,
        replace_classifiers: bool = True,
    ):
        """
        Initialize a local classifier per node.

        Parameters
        ----------
        local_classifier : BaseEstimator, default=LogisticRegression
            The local_classifier used to create the collection of local classifiers. Needs to have fit, predict and
            clone methods.
        binary_policy : str, default="siblings"
            Rules for defining positive and negative training examples.
        verbose : int, default=0
            Controls the verbosity when fitting and predicting.
            See https://verboselogs.readthedocs.io/en/latest/readme.html#overview-of-logging-levels
            for more information.
        edge_list : str, default=None
            Path to write the hierarchy built.
        replace_classifiers : bool, default=True
            Turns on (True) the replacement of a local classifier with a constant classifier when trained on only
            a single unique class.
        """
        self.local_classifier = local_classifier
        self.binary_policy = binary_policy
        self.verbose = verbose
        self.edge_list = edge_list
        self.replace_classifiers = replace_classifiers

    def fit(self, X, y):
        """
        Fit a local classifier per node.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, its dtype will be converted
            to ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csc_matrix``.
        y : array-like of shape (n_samples, n_levels)
            The target values, i.e., hierarchical class labels for classification.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        # Check that X and y have correct shape
        # and convert them to np.ndarray if need be
        self.X_, self.y_ = check_X_y(X, y, multi_output=True, accept_sparse="csr")

        # Create and configure logger
        self._create_logger()

        # Avoids creating more columns in prediction if edges are a->b and b->c,
        # which would generate the prediction a->b->c
        self._disambiguate()

        # Create DAG from self.y_ and store to self.hierarchy_
        self._create_digraph()

        # If user passes edge_list, then export
        # DAG to CSV file to visualize with Gephi
        self._export_digraph()

        # Assert that graph is directed acyclic
        self._assert_digraph_is_dag()

        # If y is 1D, convert to 2D for binary policies
        self._convert_1d_y_to_2d()

        # Initialize policy
        self._initialize_binary_policy()

        # Detect root(s) and add artificial root to DAG
        self._add_artificial_root()

        # Initialize local classifiers in DAG
        self._initialize_local_classifiers()

        # Fit local classifiers in DAG
        self._fit_digraph()

        # TODO: Store the classes seen during fit

        # TODO: Add function to allow user to change local classifier

        # TODO: Add parameter to receive hierarchy as parameter in constructor

        # TODO: Add support to empty labels in some levels

        # TODO: Parallelize fit

        # Delete unnecessary variables
        self._clean_up()

        # Return the classifier
        return self

    def predict(self, X):
        """
        Predict classes for the given data.

        Hierarchical labels are returned.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted
            to ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        y : ndarray of shape (n_samples,) or (n_samples, n_outputs)
            The predicted classes.
        """
        # Check if fit has been called
        check_is_fitted(self)

        # Input validation
        X = check_array(X, accept_sparse="csr")

        y = np.empty((X.shape[0], self.max_levels_), dtype=self.dtype_)

        # TODO: Add threshold to stop prediction halfway if need be

        bfs = nx.bfs_successors(self.hierarchy_, source=self.root_)

        self.logger_.info("Predicting")

        for predecessor, successors in bfs:
            if predecessor == self.root_:
                mask = [True] * X.shape[0]
                subset_x = X[mask]
            else:
                mask = np.isin(y, predecessor).any(axis=1)
                subset_x = X[mask]
            if subset_x.shape[0] > 0:
                probabilities = np.zeros((subset_x.shape[0], len(successors)))
                for i, successor in enumerate(successors):
                    successor_name = str(successor).split(self.separator_)[-1]
                    self.logger_.info(f"Predicting for node '{successor_name}'")
                    classifier = self.hierarchy_.nodes[successor]["classifier"]
                    positive_index = np.where(classifier.classes_ == 1)[0]
                    probabilities[:, i] = classifier.predict_proba(subset_x)[
                        :, positive_index
                    ][:, 0]
                highest_probability = np.argmax(probabilities, axis=1)
                prediction = []
                for i in highest_probability:
                    prediction.append(successors[i])
                level = nx.shortest_path_length(
                    self.hierarchy_, self.root_, predecessor
                )
                y[mask, level] = prediction

        # Convert back to 1D if there is only 1 column to pass all sklearn's checks
        if self.max_levels_ == 1:
            y = y.flatten()

        # Remove separator from predictions
        if y.ndim == 2:
            for i in range(y.shape[0]):
                for j in range(1, y.shape[1]):
                    y[i, j] = y[i, j].split(self.separator_)[-1]

        return y

    def _create_logger(self):
        # Create logger
        self.logger_ = logging.getLogger("LCPN")
        self.logger_.setLevel(self.verbose)

        # Create console handler and set verbose level
        ch = logging.StreamHandler()
        ch.setLevel(self.verbose)

        # Create formatter
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        # Add formatter to ch
        ch.setFormatter(formatter)

        # Add ch to logger
        self.logger_.addHandler(ch)

    def _disambiguate(self):
        self.separator_ = "::HiClass::Separator::"
        if self.y_.ndim == 2:
            new_y = []
            for i in range(self.y_.shape[0]):
                row = [self.y_[i, 0]]
                for j in range(1, self.y_.shape[1]):
                    row.append(str(row[-1]) + self.separator_ + str(self.y_[i, j]))
                new_y.append(np.asarray(row, dtype=np.str_))
            self.y_ = np.array(new_y)

    def _create_digraph(self):
        # Create DiGraph
        self.hierarchy_ = nx.DiGraph()

        # Save dtype of y_
        self.dtype_ = self.y_.dtype

        # Check dimension of labels
        if self.y_.ndim == 2:
            # 2D labels
            # Create max_levels variable
            self.max_levels_ = self.y_.shape[1]
            rows, columns = self.y_.shape
            self.logger_.info(f"Creating digraph from {rows} 2D labels")
            for row in range(rows):
                for column in range(columns - 1):
                    self.hierarchy_.add_edge(
                        self.y_[row, column], self.y_[row, column + 1]
                    )

        elif self.y_.ndim == 1:
            # 1D labels
            # Create max_levels_ variable
            self.max_levels_ = 1
            self.logger_.info(f"Creating digraph from {self.y_.size} 1D labels")
            for label in self.y_:
                self.hierarchy_.add_node(label)

        else:
            # Unsuported dimension
            self.logger_.error(f"y with {self.y_.ndim} dimensions detected")
            raise ValueError(
                f"Creating graph from y with {self.y_.ndim} dimensions is not supported"
            )

    def _convert_1d_y_to_2d(self):
        # This conversion is necessary for the binary policies
        if self.y_.ndim == 1:
            self.y_ = np.reshape(self.y_, (-1, 1))

    def _export_digraph(self):
        # Check if edge_list is set
        if self.edge_list:
            # Add quotes to all nodes in case the text has commas
            mapping = {}
            for node in self.hierarchy_:
                mapping[node] = '"{}"'.format(node.split(self.separator_)[-1])
            hierarchy = nx.relabel_nodes(self.hierarchy_, mapping, copy=True)
            # Export DAG to CSV file
            self.logger_.info(f"Writing edge list to file {self.edge_list}")
            nx.write_edgelist(hierarchy, self.edge_list, delimiter=",")

    def _assert_digraph_is_dag(self):
        # Assert that graph is directed acyclic
        if not nx.is_directed_acyclic_graph(self.hierarchy_):
            self.logger_.error("Cycle detected in graph")
            raise ValueError("Graph is not directed acyclic")

    def _initialize_binary_policy(self):
        if isinstance(self.binary_policy, str):
            self.logger_.info(f"Initializing {self.binary_policy} binary policy")
            try:
                self.binary_policy_ = BinaryPolicy.IMPLEMENTED_POLICIES[
                    self.binary_policy.lower()
                ](self.hierarchy_, self.X_, self.y_)
            except KeyError:
                self.logger_.error(
                    f"Policy {self.binary_policy} not implemented. Available policies are:\n"
                    + f"{list(BinaryPolicy.IMPLEMENTED_POLICIES.keys())}"
                )
                raise KeyError(f"Policy {self.binary_policy} not implemented.")
        else:
            self.logger_.error("Binary policy is not a string")
            raise ValueError(
                f"Binary policy type must str, not {type(self.binary_policy)}."
            )

    def _add_artificial_root(self):
        # Detect root(s)
        roots = [
            node for node, in_degree in self.hierarchy_.in_degree() if in_degree == 0
        ]
        self.logger_.info(f"Detected {len(roots)} roots")

        # Add artificial root as predecessor to root(s) detected
        self.root_ = "hiclass::root"
        for old_root in roots:
            self.hierarchy_.add_edge(self.root_, old_root)

    def _initialize_local_classifiers(self):
        # Create a deep copy of the local classifier specified
        # for each node in the hierarchy and save to attribute "classifier"
        self.logger_.info("Initializing local classifiers")
        local_classifiers = {}
        if self.local_classifier is None:
            self.local_classifier_ = LogisticRegression()
        else:
            self.local_classifier_ = self.local_classifier
        for node in self.hierarchy_.nodes:
            # Skip only root node
            if node != self.root_:
                local_classifiers[node] = {
                    "classifier": deepcopy(self.local_classifier_)
                }
        nx.set_node_attributes(self.hierarchy_, local_classifiers)

    def _fit_digraph(self):
        self.logger_.info("Fitting local classifiers")
        nodes = list(self.hierarchy_.nodes)
        # Remove root because it does not need to be fitted
        nodes.remove(self.root_)
        for index, node in enumerate(nodes):
            node_name = str(node).split(self.separator_)[-1]
            self.logger_.info(
                f"Fitting local classifier for node '{node_name}' ({index + 1}/{len(nodes)})"
            )
            classifier = self.hierarchy_.nodes[node]["classifier"]
            X, y = self.binary_policy_.get_binary_examples(node)
            unique_y = np.unique(y)
            if len(unique_y) == 1 and self.replace_classifiers:
                node_name = str(node).split(self.separator_)[-1]
                self.logger_.warning(f"Fitting ConstantClassifier for node '{node_name}'")
                self.hierarchy_.nodes[node]["classifier"] = ConstantClassifier()
                classifier = self.hierarchy_.nodes[node]["classifier"]
            classifier.fit(X, y)

    def _clean_up(self):
        self.logger_.info("Cleaning up variables that can take a lot of disk space")
        del self.X_
        del self.y_
        del self.binary_policy_
